;;; Mudsync --- Live hackable MUD
;;; Copyright © 2016 Christopher Allan Webber <cwebber@dustycloud.org>
;;;
;;; This file is part of Mudsync.
;;;
;;; Mudsync is free software; you can redistribute it and/or modify it
;;; under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; Mudsync is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with Mudsync.  If not, see <http://www.gnu.org/licenses/>.

;;; Game actor
;;; ==========

(define-module (mudsync gameobj)
  #:use-module (mudsync command)
  #:use-module (8sync systems actors)
  #:use-module (8sync agenda)
  #:use-module (srfi srfi-1)
  #:use-module (ice-9 match)
  #:use-module (oop goops)
  #:export (<gameobj>
            gameobj-simple-name-f

            gameobj-loc
            gameobj-gm
            gameobj-name
            gameobj-name-f

            gameobj-actions))

;;; Gameobj
;;; =======


;;; Actions supported by all gameobj
(define gameobj-actions
  (build-actions
   (init (wrap-apply gameobj-init))
   (get-commands (wrap-apply gameobj-get-commands))
   (get-container-commands (wrap-apply gameobj-get-container-commands))
   (get-occupants (wrap-apply gameobj-get-occupants))
   (add-occupant! (wrap-apply gameobj-add-occupant!))
   (remove-occupant! (wrap-apply gameobj-remove-occupant!))
   (set-loc! (wrap-apply gameobj-set-loc!))
   (get-name (wrap-apply gameobj-get-name))
   (get-desc (wrap-apply gameobj-get-desc))
   (goes-by (wrap-apply gameobj-goes-by))
   (visible-name (wrap-apply gameobj-visible-name))))

;;; *all* game components that talk to players should somehow
;;; derive from this class.
;;; And all of them need a GM!

(define-class <gameobj> (<actor>)
  ;; location id
  (loc #:init-value #f
       #:getter gameobj-loc)
  
  ;; Uses a hash table like a set (values ignored)
  (occupants #:init-thunk make-hash-table
             #:getter gameobj-occupants)

  ;; game master id
  (gm #:init-keyword #:gm
      #:getter gameobj-gm)
  ;; a name to be known by
  (name #:init-keyword #:name
        #:init-value #f)
  (goes-by #:init-keyword #:goes-by
           #:init-value #f)

  (desc #:init-value ""
        #:init-keyword #:desc)

  ;; how to print our name
  (name-f #:init-keyword #:name-f
          #:getter gameobj-name-f
          #:init-value (wrap gameobj-simple-name-f))

  ;; Commands we can handle
  (commands #:init-value '())

  ;; Commands we can handle by being something's container
  (container-commands #:init-value '())
  (message-handler
   #:init-value
   (simple-dispatcher gameobj-actions))

  ;; Most objects are generally visible by default
  (generally-visible #:init-value #t)
  ;; @@: Would be preferable to be using generic methods for this...
  ;;   Hopefully we can port this to Guile 2.2 soon...
  (visible-to-player?
   #:init-value (wrap-apply gameobj-visible-to-player?)))


;;; gameobj message handlers
;;; ========================

;; Kind of a useful utility, maybe?
(define (simple-slot-getter slot)
  (lambda (actor message)
    (reply-message actor message
                   #:val (slot-ref actor slot))))


;; @@: This could be kind of a messy way of doing gameobj-init
;;   stuff.  If only we had generic methods :(
(define-mhandler (gameobj-init actor message)
  "Your most basic game object init procedure.  Does nothing."
  #f)

(define (gameobj-goes-by gameobj)
  "Find the name we go by.  Defaults to #:name if nothing else provided."
  (cond ((slot-ref gameobj 'goes-by) =>
         identity)
        ((slot-ref gameobj 'name) =>
         (lambda (name)
           (list name)))
        (else '())))

(define (val-or-run val-or-proc)
  "Evaluate if a procedure, or just return otherwise"
  (if (procedure? val-or-proc)
      (val-or-proc)
      val-or-proc))

(define (filter-commands commands verb)
  (filter
   (lambda (cmd)
     (equal? (command-verbs cmd)
             verb))
   commands))

(define-mhandler (gameobj-get-commands actor message verb)
  "Get commands a co-occupant of the room might execute for VERB"
  (define filtered-commands
    (filter-commands (val-or-run (slot-ref actor 'commands))
                     verb))
  (<-reply actor message
           #:commands filtered-commands
           #:goes-by (gameobj-goes-by actor)))

(define-mhandler (gameobj-get-container-commands actor message verb)
  "Get commands as the container / room of message's sender"
  (define filtered-commands
    (filter-commands (val-or-run (slot-ref actor 'container-commands))
                     verb))
  (<-reply actor message #:commands filtered-commands))

(define-mhandler (gameobj-add-occupant! actor message who)
  "Add an actor to our list of present occupants"
  (hash-set! (slot-ref actor 'occupants)
             who #t))

(define-mhandler (gameobj-remove-occupant! actor message who)
  "Remove an occupant from the room."
  (hash-remove! (slot-ref actor 'occupants) who))

(define-mhandler (gameobj-get-occupants actor message)
  "Get all present occupants of the room."
  (define occupants
    (hash-map->list (lambda (key val) key)
                    (gameobj-occupants actor)))

  (<-reply actor message
           #:occupants occupants))

;; @@: Should it really be #:id ?  Maybe #:loc-id or #:loc?
(define-mhandler (gameobj-set-loc! actor message loc)
  "Set the location of this object."
  (define old-loc (gameobj-loc actor))
  (format #t "DEBUG: Location set to ~s for ~s\n"
          loc (actor-id-actor actor))

  (slot-set! actor 'loc loc)
  ;; Change registation of where we currently are
  (if loc
      (<-wait actor loc 'add-occupant! #:who (actor-id actor)))
  (if old-loc
      (<-wait actor old-loc 'remove-occupant! #:who (actor-id actor))))

(define gameobj-get-name (simple-slot-getter 'name))
(define gameobj-get-desc (simple-slot-getter 'desc))

(define (gameobj-simple-name-f gameobj)
  "Simplest version: return ourselves for our name."
  (gameobj-name gameobj))

(define (gameobj-visible-to-player? gameobj whos-looking)
  "Check to see whether we're visible to the player or not.
By default, this is whether or not the generally-visible flag is set."
  (slot-ref gameobj 'generally-visible))

(define-mhandler (gameobj-visible-name actor message whos-looking)
  ;; Are we visible?
  (define we-are-visible
    ((slot-ref actor 'visible-to-player?) actor whos-looking))

  (define name-to-return
    (if we-are-visible
        ;; Return our name
        (match (slot-ref actor 'name)
          ((? procedure? name-proc)
           (name-proc actor whos-looking))
          ((? string? name)
           name)
          (#f #f))
        #f))
  (<-reply actor message #:text name-to-return))
